@@ -13,10 +13,12 @@ import warnings
 
 from contextlib import contextmanager
 from io import BufferedReader, open
+from typing import Union
 
 from pyfatfs import FAT_OEM_ENCODING
 from pyfatfs.EightDotThree import EightDotThree
 from pyfatfs.FATDirectoryEntry import FATDirectoryEntry, FATLongDirectoryEntry
+from pyfatfs.FATHeader import FATHeader, FAT32Header, FAT12Header
 from pyfatfs._exceptions import PyFATException, NotAFatEntryException
 
 
@@ -100,29 +102,17 @@ class PyFat(object):
                        "BPB_FATSz16", "BPB_SecPerTrk", "BPB_NumHeads",
                        "BPB_HiddSec", "BPB_TotSec32"]
 
-    #: FAT12/16 header layout in struct formatted string
-    fat12_header_layout = "<BBBL11s8s"
-    #: FAT12/16 header fields when extracted with fat12_header_layout
-    fat12_header_vars = ["BS_DrvNum", "BS_Reserved1", "BS_BootSig",
-                         "BS_VolID", "BS_VolLab", "BS_FilSysType"]
-
-    #: FAT32 header layout in struct formatted string
-    fat32_header_layout = "<LHHLHH12sBBBL11s8s"
-    #: FAT32 header fields when extracted with fat32_header_layout
-    fat32_header_vars = ["BPB_FATSz32", "BPB_ExtFlags", "BPB_FSVer",
-                         "BPB_RootClus", "BPB_FSInfo", "BPB_BkBootSec",
-                         "BPB_Reserved", "BS_DrvNum", "BS_Reserved1",
-                         "BS_BootSig", "BS_VolID", "BS_VolLab",
-                         "BS_FilSysType"]
-
-    #: FAT12/16 bit mask for clean shutdown bit
-    FAT12_CLEAN_SHUTDOWN_BIT_MASK = 0x8000
-    #: FAT12/16 bit mask for volume error bit
-    FAT12_DRIVE_ERROR_BIT_MASK = 0x4000
+    #: FAT16 bit mask for clean shutdown bit
+    FAT16_CLEAN_SHUTDOWN_BIT_MASK = 0x8000
+    #: FAT16 bit mask for volume error bit
+    FAT16_DRIVE_ERROR_BIT_MASK = 0x4000
     #: FAT32 bit mask for clean shutdown bit
-    FAT32_CLEAN_SHUTDOWN_BIT_MASK = 0x0800000
+    FAT32_CLEAN_SHUTDOWN_BIT_MASK = 0x8000000
     #: FAT32 bit mask for volume error bit
-    FAT32_DRIVE_ERROR_BIT_MASK = 0x0400000
+    FAT32_DRIVE_ERROR_BIT_MASK = 0x4000000
+
+    #: Dirty bit in FAT header
+    FAT_DIRTY_BIT_MASK = 0x01
 
     def __init__(self,
                  encoding: str = 'ibm437',
@@ -136,17 +126,18 @@ class PyFat(object):
         """
         self.__fp = None
         self.__fp_offset = offset
+        self._fat_size = 0
         self.bpb_header = None
-        self.fat_header = None
+        self.fat_header: FATHeader = FATHeader()
         self.root_dir = None
         self.root_dir_sector = 0
         self.root_dir_sectors = 0
         self.bytes_per_cluster = 0
         self.first_data_sector = 0
+        self.first_free_cluster = 0
         self.fat_type = self.FAT_TYPE_UNKNOWN
         self.fat = {}
         self.initialised = False
-        self.fat_clusterchains = {}
         self.encoding = encoding
         self.is_read_only = True
         self.__lock = threading.Lock()
@@ -177,6 +168,63 @@ class PyFat(object):
             self.__seek(cluster_address)
             return self.__fp.read(sz)
 
+    def __get_clean_shutdown_bitmask(self):
+        """Get clean shutdown bitmask for current FS.
+
+        :raises: AttributeError
+        """
+        return getattr(self, f"FAT{self.fat_type}_CLEAN_SHUTDOWN_BIT_MASK")
+
+    def _is_dirty(self) -> bool:
+        """Check whether or not the partition currently is dirty."""
+        try:
+            clean_shutdown_bitmask = self.__get_clean_shutdown_bitmask()
+        except AttributeError:
+            # Bit not set on FAT12
+            dos_dirty = False
+        else:
+            dos_dirty = (self.fat[1] &
+                         clean_shutdown_bitmask) != clean_shutdown_bitmask
+
+        nt_dirty = (self.fat_header["BS_Reserved1"] &
+                    self.FAT_DIRTY_BIT_MASK) == self.FAT_DIRTY_BIT_MASK
+
+        return dos_dirty or nt_dirty
+
+    def _mark_dirty(self):
+        """Mark partition as not cleanly unmounted.
+
+        Apparently the dirty bit in FAT[1] is used by DOS,
+        while BS_Reserved1 is used by NT. Always set both.
+        """
+        try:
+            clean_shutdown_bitmask = self.__get_clean_shutdown_bitmask()
+        except AttributeError:
+            pass
+        else:
+            # Only applicable for FAT16/32
+            self.fat[1] = (self.fat[1] & ~clean_shutdown_bitmask) | \
+                          (0 & clean_shutdown_bitmask)
+            self.flush_fat()
+
+        self.fat_header["BS_Reserved1"] |= self.FAT_DIRTY_BIT_MASK
+        self._write_fat_header()
+
+    def _mark_clean(self):
+        """Mark partition as cleanly unmounted."""
+        try:
+            clean_shutdown_bitmask = self.__get_clean_shutdown_bitmask()
+        except AttributeError:
+            pass
+        else:
+            self.fat[1] |= clean_shutdown_bitmask
+            self.flush_fat()
+
+        self.fat_header["BS_Reserved1"] = (self.fat_header["BS_Reserved1"]
+                                           & ~self.FAT_DIRTY_BIT_MASK) | \
+                                          (0 & self.FAT_DIRTY_BIT_MASK)
+        self._write_fat_header()
+
     def open(self, filename: str, read_only: bool = False):
         """Open filesystem for usage with PyFat.
 
@@ -201,6 +249,13 @@ class PyFat(object):
         # Parse FAT
         self._parse_fat()
 
+        # Check for clean shutdown
+        if self._is_dirty():
+            warnings.warn("Filesystem was not cleanly unmounted on last "
+                          "access. Check for data corruption.")
+        if not self.is_read_only:
+            self._mark_dirty()
+
         # Parse root directory
         # TODO: Inefficient to always recursively parse the root dir.
         #       It would make sense to parse it on demand instead.
@@ -220,13 +275,12 @@ class PyFat(object):
         return self.bpb_header["BPB_TotSec32"]
 
     def _get_fat_size_count(self):
+        """Get BPB_FATsz value."""
         if self.bpb_header["BPB_FATSz16"] != 0:
             return self.bpb_header["BPB_FATSz16"]
 
-        # Only possible with FAT32
-        self.__parse_fat32_header()
         try:
-            return self.fat_header["BPB_FATSz32"]
+            return self._parse_fat_header(force_fat32=True)["BPB_FATSz32"]
         except KeyError:
             raise PyFATException("Invalid FAT size of 0 detected in header, "
                                  "cannot continue")
@@ -236,7 +290,7 @@ class PyFat(object):
         """Parse information in FAT."""
         # Read all FATs
         fat_size = self.bpb_header["BPB_BytsPerSec"]
-        fat_size *= self._get_fat_size_count()
+        fat_size *= self._fat_size
 
         # Seek FAT entries
         first_fat_bytes = self.bpb_header["BPB_RsvdSecCnt"]
@@ -258,8 +312,7 @@ class PyFat(object):
         self.bytes_per_cluster = self.bpb_header["BPB_BytsPerSec"] * \
             self.bpb_header["BPB_SecPerClus"]
 
-        if len(fats[0]) != self.bpb_header["BPB_BytsPerSec"] * \
-                self._get_fat_size_count():
+        if len(fats[0]) != self.bpb_header["BPB_BytsPerSec"] * self._fat_size:
             raise PyFATException("Invalid length of FAT")
 
         # FAT12: 12 bits (1.5 bytes) per FAT entry
@@ -277,7 +330,13 @@ class PyFat(object):
 
             if self.fat_type == self.FAT_TYPE_FAT12:
                 fat_nibble = fats[0][int(curr):math.ceil(offset)]
-                self.fat[cluster] = struct.unpack("<H", fat_nibble)[0]
+                fat_nibble = fat_nibble.ljust(2, b"\0")
+                try:
+                    self.fat[cluster] = struct.unpack("<H", fat_nibble)[0]
+                except IndexError:
+                    # Out of bounds, FAT size is not cleanly divisible by 3
+                    # Do not touch last clusters
+                    break
 
                 if cluster % 2 == 0:
                     # Even: Keep low 12-bits of word
@@ -313,25 +372,39 @@ class PyFat(object):
                                  "report this error.")
 
     @_init_check
-    def byte_repr(self) -> bytearray:
+    def __bytes__(self):
         """Represent current state of FAT as bytes.
 
         :returns: `bytes` representation of FAT.
         """
-        b = bytearray(b'')
+        b = b''
         if self.fat_type == self.FAT_TYPE_FAT12:
-            raise NotImplementedError("FAT12 write support currently "
-                                      "not implemented!")
-        elif self.fat_type == self.FAT_TYPE_FAT16:
-            fmt = "<H"
-        else:
-            # FAT32
-            fmt = "<L"
+            for i, e in enumerate(self.fat):
+                if i % 2 == 0:
+                    b += struct.pack("<H", e)
+                else:
+                    nibble = b[-1:]
+                    nibble = struct.unpack("<B", nibble)[0]
+                    b = b[:-1]
+                    b += struct.pack("<BB", ((e & 0xF) << 4) | nibble, e >> 4)
 
-        for c in self.fat:
-            b += struct.pack(fmt, c)
+        else:
+            if self.fat_type == self.FAT_TYPE_FAT16:
+                fmt = "H"
+            else:
+                # FAT32
+                fmt = "L"
+
+            b = struct.pack(f"<{fmt * len(self.fat)}",
+                            *self.fat)
         return b
 
+    def byte_repr(self) -> bytearray:
+        """Do not use."""
+        warnings.warn("byte_repr is deprecated, directly cast "
+                      "to bytes instead", DeprecationWarning)
+        return bytearray(bytes(self))
+
     @_init_check
     @_readonly_check
     def _write_data_to_address(self, data: bytes,
@@ -360,6 +433,7 @@ class PyFat(object):
             tmp_fat = self.fat.copy()
             for cl in self.get_cluster_chain(cluster):
                 tmp_fat[cl] = _freeclus
+                self.first_free_cluster = min(cl, self.first_free_cluster)
             self.fat = tmp_fat
 
     @_init_check
@@ -385,6 +459,8 @@ class PyFat(object):
         for c in self.get_cluster_chain(cluster):
             cluster_sz += self.bytes_per_cluster
             last_cluster = c
+            if cluster_sz >= data_sz:
+                break
 
         if data_sz > cluster_sz:
             if extend_cluster is False:
@@ -398,9 +474,8 @@ class PyFat(object):
 
         # Fill rest of data with zeroes if erase is set to True
         if erase:
-            new_sz = self.bytes_per_cluster - data_sz
-            new_sz %= self.bytes_per_cluster
-            new_sz += data_sz
+            new_sz = max(1, math.ceil(data_sz / self.bytes_per_cluster))
+            new_sz *= self.bytes_per_cluster
             data += b'\0' * (new_sz - data_sz)
 
         # Write actual data
@@ -410,21 +485,24 @@ class PyFat(object):
             t = bytes_written
             bytes_written += self.bytes_per_cluster
             self._write_data_to_address(data[t:bytes_written], b)
+            if bytes_written >= len(data):
+                break
 
     @_init_check
     @_readonly_check
     def flush_fat(self) -> None:
         """Flush FAT(s) to disk."""
         fat_size = self.bpb_header["BPB_BytsPerSec"]
-        fat_size *= self._get_fat_size_count()
+        fat_size *= self._fat_size
 
         first_fat_bytes = self.bpb_header["BPB_RsvdSecCnt"]
         first_fat_bytes *= self.bpb_header["BPB_BytsPerSec"]
 
-        for i in range(self.bpb_header["BPB_NumFATS"]):
-            with self.__lock:
+        with self.__lock:
+            binary_fat = bytes(self)
+            for i in range(self.bpb_header["BPB_NumFATS"]):
                 self.__seek(first_fat_bytes + (i * fat_size))
-                self.__fp.write(self.byte_repr())
+                self.__fp.write(binary_fat)
 
     def calc_num_clusters(self, size: int = 0) -> int:
         """Calculate the number of required clusters.
@@ -454,8 +532,8 @@ class PyFat(object):
 
         # Fill list of found free clusters
         free_clusters = []
-        for i, _ in enumerate(self.fat):
-            if min_clus > i > max_clus:
+        for i in range(self.first_free_cluster, len(self.fat)):
+            if min_clus > i or i > max_clus:
                 # Ignore out of bound entries
                 continue
 
@@ -479,6 +557,7 @@ class PyFat(object):
             raise PyFATException(f"Not enough free space to allocate "
                                  f"{size} bytes ({free_space} bytes free)",
                                  errno=errno.ENOSPC)
+        self.first_free_cluster = i
 
         # Allocate cluster chain in FAT
         eoc_max = self.FAT_CLUSTER_VALUES[self.fat_type]["END_OF_CLUSTER_MAX"]
@@ -518,7 +597,7 @@ class PyFat(object):
         dir_entries = b''
         d, f, s = dir_entry.get_entries()
         for d in list(itertools.chain(d, f, s)):
-            dir_entries += d.byte_repr()
+            dir_entries += bytes(d)
 
         # Write content
         if not is_root_dir or self.fat_type == self.FAT_TYPE_FAT32:
@@ -633,8 +712,11 @@ class PyFat(object):
                                      address: int = 0,
                                      max_address: int = 0,
                                      tmp_lfn_entry: FATLongDirectoryEntry =
-                                     FATLongDirectoryEntry()):
+                                     None):
         """Parse directory entries in address range."""
+        if tmp_lfn_entry is None:
+            tmp_lfn_entry = FATLongDirectoryEntry()
+
         dir_hdr_size = FATDirectoryEntry.FAT_DIRECTORY_HEADER_SIZE
 
         if max_address == 0:
@@ -646,6 +728,7 @@ class PyFat(object):
             # Parse each entry
             dir_hdr = self.__parse_dir_entry(hdr_addr)
             dir_sn = EightDotThree(encoding=self.encoding)
+            dir_first_byte = dir_hdr["DIR_Name"][0]
             try:
                 dir_sn.set_byte_name(dir_hdr["DIR_Name"])
             except NotAFatEntryException as ex:
@@ -661,7 +744,7 @@ class PyFat(object):
                 dir_hdr["DIR_Name"] = dir_sn
 
             # Long File Names
-            if FATLongDirectoryEntry.is_lfn_entry(dir_hdr["DIR_Name"],
+            if FATLongDirectoryEntry.is_lfn_entry(dir_first_byte,
                                                   dir_hdr["DIR_Attr"]):
                 self.parse_lfn_entry(tmp_lfn_entry, hdr_addr)
                 continue
@@ -757,7 +840,7 @@ class PyFat(object):
     def close(self):
         """Close session and free up all handles."""
         if not self.is_read_only:
-            self.flush_fat()
+            self._mark_clean()
 
         self.__fp.close()
         self.initialised = False
@@ -769,8 +852,10 @@ class PyFat(object):
         except PyFATException:
             pass
 
-    def __determine_fat_type(self):
-        """Determine FAT size.
+    def __determine_fat_type(self) -> Union["PyFat.FAT_TYPE_FAT12",
+                                            "PyFat.FAT_TYPE_FAT16",
+                                            "PyFat.FAT_TYPE_FAT32"]:
+        """Determine FAT type.
 
         An internal method to determine whether this volume is FAT12,
         FAT16 or FAT32.
@@ -784,39 +869,67 @@ class PyFat(object):
             total_sectors = self.bpb_header["BPB_TotSec32"]
 
         rsvd_sectors = self.bpb_header["BPB_RsvdSecCnt"]
-        fat_sz = self.bpb_header["BPB_NumFATS"] * self._get_fat_size_count()
+        fat_sz = self.bpb_header["BPB_NumFATS"] * self._fat_size
         root_dir_sectors = self.root_dir_sectors
         data_sec = total_sectors - (rsvd_sectors + fat_sz + root_dir_sectors)
         count_of_clusters = data_sec // self.bpb_header["BPB_SecPerClus"]
 
         if count_of_clusters < 4085:
-            fat_type = self.FAT_TYPE_FAT12
+            msft_fat_type = self.FAT_TYPE_FAT12
         elif count_of_clusters < 65525:
-            fat_type = self.FAT_TYPE_FAT16
+            msft_fat_type = self.FAT_TYPE_FAT16
         else:
-            fat_type = self.FAT_TYPE_FAT32
+            msft_fat_type = self.FAT_TYPE_FAT32
 
-        return fat_type
+        if self.bpb_header["BPB_FATSz16"] == 0:
+            fat_hdr = self._parse_fat_header(force_fat32=True)
+            if fat_hdr["BPB_FATSz32"] != 0:
+                linux_fat_type = self.FAT_TYPE_FAT32
+            else:
+                linux_fat_type = msft_fat_type
+        elif count_of_clusters >= 4085:
+            linux_fat_type = self.FAT_TYPE_FAT16
+        else:
+            linux_fat_type = self.FAT_TYPE_FAT12
 
-    def __parse_fat12_header(self):
-        """Parse FAT12/16 header."""
+        if msft_fat_type != linux_fat_type:
+            warnings.warn(f"Unable to reliably determine FAT type, "
+                          f"guessing either FAT{msft_fat_type} or "
+                          f"FAT{linux_fat_type}. Opting for "
+                          f"FAT{linux_fat_type}.")
+        return linux_fat_type
+
+    @_readonly_check
+    def _write_fat_header(self):
+        with self.__lock:
+            self.__seek(36)
+            self.__fp.write(bytes(self.fat_header))
+
+    def _parse_fat_header(self, force_fat32: bool = False) -> FATHeader:
+        """Parse FAT header."""
         with self.__lock:
             self.__seek(0)
             boot_sector = self.__fp.read(512)
 
-        header = struct.unpack(self.fat12_header_layout,
-                               boot_sector[36:][:26])
-        self.fat_header = dict(zip(self.fat12_header_vars, header))
+        if self.fat_type in [self.FAT_TYPE_FAT12, self.FAT_TYPE_FAT16] \
+                and not force_fat32:
+            fat_header = FAT12Header()
+            fat_header.parse_header(boot_sector[36:][:26])
+            return fat_header
+        elif self.fat_type == self.FAT_TYPE_FAT32 or force_fat32:
+            # FAT32, probably - probe for it
+            if self.bpb_header["BPB_FATSz16"] != 0:
+                raise PyFATException(f"Invalid BPB_FATSz16 value of "
+                                     f"'{self.bpb_header['BPB_FATSz16']}', "
+                                     f"filesystem corrupt?",
+                                     errno=errno.EINVAL)
 
-    def __parse_fat32_header(self):
-        """Parse FAT32 header."""
-        with self.__lock:
-            self.__seek(0)
-            boot_sector = self.__fp.read(512)
+            fat_header = FAT32Header()
+            fat_header.parse_header(boot_sector[36:][:54])
+            return fat_header
 
-        header = struct.unpack(self.fat32_header_layout,
-                               boot_sector[36:][:54])
-        self.fat_header = dict(zip(self.fat32_header_vars, header))
+        raise PyFATException("Unknown FAT filesystem type, "
+                             "filesystem corrupt?", errno=errno.EINVAL)
 
     def parse_header(self):
         """Parse BPB & FAT headers in opened file."""
@@ -830,6 +943,13 @@ class PyFat(object):
         # Verify BPB headers
         self.__verify_bpb_header()
 
+        # Determine FAT type
+        self._fat_size = self._get_fat_size_count()
+        self.fat_type = self.__determine_fat_type()
+
+        # Parse FAT type specific header
+        self.fat_header = self._parse_fat_header()
+
         # Calculate root directory sectors and starting point of root directory
         root_entries = self.bpb_header["BPB_RootEntCnt"]
         hdr_size = FATDirectoryEntry.FAT_DIRECTORY_HEADER_SIZE
@@ -839,36 +959,11 @@ class PyFat(object):
 
         self.root_dir_sectors = ((root_entries * hdr_size) +
                                  (bytes_per_sec - 1)) // bytes_per_sec
-        self.root_dir_sector = rsvd_secs + (self._get_fat_size_count() *
-                                            num_fats)
+        self.root_dir_sector = rsvd_secs + (self._fat_size * num_fats)
 
         # Calculate first data sector
-        self.first_data_sector = rsvd_secs + (num_fats *
-                                              self._get_fat_size_count()) + \
-            self.root_dir_sectors
-
-        # Determine FAT type
-        self.fat_type = self.__determine_fat_type()
-
-        # Parse FAT type specific header
-        if self.fat_type in [self.FAT_TYPE_FAT12, self.FAT_TYPE_FAT16]:
-            self.__parse_fat12_header()
-
-            self.__verify_fat12_header()
-        elif self.fat_type == self.FAT_TYPE_FAT32:
-            # FAT32, probably - probe for it
-            if self.bpb_header["BPB_FATSz16"] != 0:
-                raise PyFATException(f"Invalid BPB_FATSz16 value of "
-                                     f"'{self.bpb_header['BPB_FATSz16']}', "
-                                     f"filesystem corrupt?",
-                                     errno=errno.EINVAL)
-
-            self.__parse_fat32_header()
-
-            # TODO: Verify FAT32 header
-        else:
-            raise PyFATException("Unknown FAT filesystem type, "
-                                 "filesystem corrupt?", errno=errno.EINVAL)
+        self.first_data_sector = (rsvd_secs + (num_fats * self._fat_size) +
+                                  self.root_dir_sectors)
 
         # Check signature
         with self.__lock:
@@ -881,23 +976,6 @@ class PyFat(object):
         # Initialisation finished
         self.initialised = True
 
-    def __verify_fat12_header(self):
-        """Verify FAT12/16 header for correctness."""
-        # Extended boot signature had two different versions:
-        # 0x00: Pre-DOS 4.0 BPB (no EBPB present)
-        # 0x28: Only BS_VolID is present
-        # 0x29: Full FAT12 bootsector present
-        bs_bootsig = self.fat_header["BS_BootSig"]
-        if bs_bootsig == 0x29:
-            bs_filsystype = self.fat_header["BS_FilSysType"]
-            if bs_filsystype not in (self.FS_TYPES[self.FAT_TYPE_UNKNOWN],
-                                     self.FS_TYPES[self.fat_type]):
-                print(f"BS_FilSysType contains {bs_filsystype}, unexpected "
-                      f"for FAT{self.fat_type}")
-        elif bs_bootsig not in (0x28, 0x00, 0x2a):
-            raise PyFATException("Invalid FAT extended boot signature value "
-                                 "encountered: '{}'".format(hex(bs_bootsig)))
-
     def __verify_bpb_header(self):
         """Verify BPB header for correctness."""
         if self.bpb_header["BS_jmpBoot"][0] == 0xEB:
