@@ -287,7 +287,7 @@ class PyFatFS(FS):
 
         # Create LFN entry if required
         _sfn = short_name.get_unpadded_filename()
-        if _sfn != dirname.upper() or self.preserve_case:
+        if _sfn != dirname.upper() or (_sfn != dirname and self.preserve_case):
             lfn_entry = make_lfn_entry(dirname, short_name)
             newdir.set_lfn_entry(lfn_entry)
 
@@ -385,7 +385,10 @@ class PyFatFS(FS):
         for d in dirs:
             self.removetree(posixpath.join(dir_path, str(d)))
 
-        return self.removedir(dir_path)
+        try:
+            self.removedir(dir_path)
+        except RemoveRootError:
+            pass
 
     def remove(self, path: str):
         """Remove a file from the filesystem.
@@ -436,7 +439,7 @@ class PyFatFS(FS):
         returns: `BinaryIO` stream
         """
         path = self.validatepath(path)
-        mode = Mode(mode)
+        mode = Mode(mode + 'b')
         if mode.create:
             if mode.exclusive:
                 try:
